@@ -1,5 +1,5 @@
 /*  GNU ddrescue - Data recovery tool
-    Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010
+    Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011
     Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
@@ -101,6 +101,13 @@ public:
     { if( block_vector.size() ) return block_vector[0].pos(); else return 0; }
 
   long long size() const throw()
+    {
+    if( block_vector.size() )
+      return block_vector.back().end() - block_vector[0].pos();
+    else return 0;
+    }
+
+  long long in_size() const throw()
     {
     long long s = 0;
     for( unsigned int i = 0; i < block_vector.size(); ++i )
@@ -108,6 +115,10 @@ public:
     return s;
     }
 
+  long long end() const throw()
+    { if( block_vector.size() ) return block_vector.back().end();
+      else return 0; }
+
   bool operator<( const Block & b ) const throw()
     { return ( block_vector.size() && block_vector.back().end() <= b.pos() ); }
 
